#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <asm/spinlock.h>
#include <asm/io.h>
#include <asm/bitops.h>
#include "pemu.h"

#define TIMEOUT      10
#define FAST_TIMEOUT  1


static volatile int irq_count;
static int running;
static int trace = 0;
static int debug = 0;
static int rcount = 0;
static char log_line[100];
static int log_len;
static unsigned short wstatus = 0xffff;
static atomic_t buffer_full = {0};

static int pemu_reschedule(struct pemu *pe);
static int pemu_wait_for(struct pemu *pe, long flags, long abort_flags,
			 unsigned long timeout, const char *abort_string);

static void
pemu_reset_ports(void)
{
	outb(MODE_PS2 | ERRINT_DISABLE | DMA_DISABLE | SERVINT_DISABLE, ECONTROL(0));
	outb(DIRECTION_IN | ACKINT_DISABLE, CONTROL(0));

	outb(MODE_PS2 | ERRINT_DISABLE | DMA_DISABLE | SERVINT_DISABLE, ECONTROL(1));
	outb(DIRECTION_IN | ACKINT_DISABLE, CONTROL(1));

	SET(Busy(H) | nAck(H) | PError(L) | Select(H) | nFault(H));
}

int
ppstatus_read_proc(char *page, char **start, off_t off,
		   int count, int *eof, void *handle)
{
	unsigned data0     = inb(DATA(0));
	unsigned status0   = inb(STATUS(0));
	unsigned control0  = inb(CONTROL(0));
	unsigned econtrol0 = inb(ECONTROL(0));
	unsigned data1     = inb(DATA(1));
	unsigned status1   = inb(STATUS(1));
	unsigned control1  = inb(CONTROL(1));
	unsigned econtrol1 = inb(ECONTROL(1));

	int len = sprintf(page, "%2.2x %2.2x %2.2x %2.2x   %2.2x %2.2x %2.2x %2.2x   %4.4x\n",
		       data0, status0, control0, econtrol0,
		       data1, status1, control1, econtrol1,
		       debug);

#if 0
	len += sprintf(page + len, "Busy(%s) | ",      CHECKV(control0, control1, Busy(H))      ? "H" : "L");
	len += sprintf(page + len, "nAck(%s) | ",      CHECKV(control0, control1, nAck(H))      ? "H" : "L");
	len += sprintf(page + len, "PError(%s) | ",    CHECKV(control0, control1, PError(H))    ? "H" : "L");
	len += sprintf(page + len, "Select(%s) | ",    CHECKV(control0, control1, Select(H))    ? "H" : "L");
	len += sprintf(page + len, "nFault(%s) | ",    CHECKV(control0, control1, nFault(H))    ? "H" : "L");
	len += sprintf(page + len, "nSelectIn(%s) | ", CHECKV(status0,  status1,  nSelectIn(H)) ? "H" : "L");
	len += sprintf(page + len, "nInit(%s) | ",     CHECKV(status0,  status1,  nInit(H))     ? "H" : "L");
	len += sprintf(page + len, "nAutoFd(%s) | ",   CHECKV(status0,  status1,  nAutoFd(H))   ? "H" : "L");
	len += sprintf(page + len, "nStrobe(%s)\n",    CHECKV(status0,  status1,  nStrobe(H))   ? "H" : "L");
#endif

	*start = 0;
	return len <= count ? len : count;
}

void
pemu_log_ppstatus(struct pemu *pe)
{
	char line[1024], *dummy;
	int len = ppstatus_read_proc(line, &dummy, 0, sizeof(line), NULL, NULL);
	pemu_log(pe->log_low, line, len);
}

static void
pemu_spp_mode(struct pemu *pe)
{
#define INBUFSIZE1 2
#define INBUFSIZE2 8
#define INBUFSIZE  (INBUFSIZE1 + INBUFSIZE2)
#define GUESSED_STROBE_POSITION 2
	unsigned short result;
	unsigned short inbuf[INBUFSIZE];
	unsigned short *p = inbuf;
	int i, count, strobe;

	unsigned char tmp = inb(CONTROL(0)) & HW_OUT_MASK;
	unsigned char control0_1 = DIRECTION_IN | ACKINT_ENABLE  | GETV(0, tmp, Busy(L) | nAck(H));
	unsigned char control0_2 = DIRECTION_IN | ACKINT_ENABLE  | GETV(0, tmp, Busy(H) | nAck(H));
	unsigned char control0_3 = DIRECTION_IN | ACKINT_DISABLE | GETV(0, tmp, Busy(H) | nAck(L));
	unsigned char control0_4 = DIRECTION_IN | ACKINT_DISABLE | GETV(0, tmp, Busy(H) | nAck(H));

	irq_count = 0;

	/* scan data & status registers for a short time */
	/* unrolled loops for speed */
	outb(control0_1, CONTROL(0));

	REPEAT(INBUFSIZE1, *p++ = inw(WSTATUS(0)); )

	outb(control0_2, CONTROL(0));

	REPEAT(INBUFSIZE2, *p++ = inw(WSTATUS(0)); )

	/* analyse scan of data & status registers */
	strobe = count = 0;
	result = inbuf[GUESSED_STROBE_POSITION];
	for (i = 0; i < INBUFSIZE; i++)
		if (i == 0 || inbuf[i] != wstatus) {
			wstatus = inbuf[i];
			count++;
			if (!strobe && CHECKV(wstatus >> 8, 0, nStrobe(L))) {
				strobe = count;
				result = wstatus;
			}
		}

	if (strobe == 1 && CHECKV(inbuf[INBUFSIZE-1] >> 8, 0, nStrobe(L)))
		strobe = 0;

	if (irq_count && strobe == 0)
		strobe = 1;

	if (!strobe)
		return;

	outb(control0_3, CONTROL(0));
	outb(control0_4, CONTROL(0));

rcount++;

	if (pe->receive)
		pe->receive(pe, result & 0xff, result >> 8);

	pemu_log_int(pe->log_low, "-> %2.2x", result & 0xff);
	pemu_log_int(pe->log_low, " %2.2x", result >> 8);
	pemu_log_int(pe->log_low, " %2.2x", inb(STATUS(1)));
#if 0
	if (irq_count != 1 || strobe > 2 || count > strobe + 1)
		/* Something strange happned, log all */
#endif
	{
		pemu_log_int(pe->log_low, "  !%d", irq_count);
		for (i = 0; i < INBUFSIZE; i++)
			if (i == 0 || inbuf[i] != wstatus) {
				wstatus = inbuf[i];
				log_len = sprintf(log_line, "  %2.2x %4.4x", i, wstatus);
				pemu_log(pe->log_low, log_line, log_len);
			}
	}
	pemu_log_int(pe->log_low, "   (#%8.8x)\n", rcount);
}


static int
pemu_send_nibble(struct pemu *pe, unsigned nibble, int eof)
{
	if (pemu_wait_for(pe, HostBusy(L),
			  Active1284(H) , TIMEOUT, "NIBBLE 1"))
		return -EINTR;

	SET(((nibble & 8) ? PtrBusy(H)    : PtrBusy(L))    |
	    ((nibble & 4) ? AckDataReq(H) : AckDataReq(L)) |
	    ((nibble & 2) ? Xflag(H)      : Xflag(L))      |
	    ((nibble & 1) ? nDataAvail(H) : nDataAvail(L)));
	SET( PtrClk(L) );

	if (pemu_wait_for(pe, HostBusy(H),
			  Active1284(H) , TIMEOUT, "NIBBLE 2"))
		return -EINTR;

	if (!eof)
		SET( PtrBusy(L) | AckDataReq(L) | Xflag(L) | nDataAvail(L) );
	else
		SET( PtrBusy(L) | AckDataReq(H) | Xflag(H) | nDataAvail(H) );
	SET(PtrClk(H));

	return 0;
}

static int
pemu_send_ecp_byte(struct pemu *pe, unsigned data, int eof)
{
	if (pemu_wait_for(pe, 0,
			  HostAck(L) |
			  Active1284(H) | nAckReverse(L) | HostClk(H) | nReverseRequest(L),
			  TIMEOUT, "ECP 0"))
		goto abort;

	SETC(0, DIRECTION_OUT | ACKINT_DISABLE,
	     PeriphClk(H) | PeriphAck(H) | nAckReverse(L) | nPeriphRequest(L) | Xflag(H));

	outb(data, DATA(0));

	SET( PeriphClk(L) );

	if (pemu_wait_for(pe, HostAck(H),
			  Active1284(H) | nAckReverse(L) | HostClk(H) | nReverseRequest(L),
			  FAST_TIMEOUT, "ECP 1"))
		goto abort;

	SET( PeriphClk(H) );

	if (pemu_wait_for(pe, HostAck(L),
			  Active1284(H) | nAckReverse(L) | HostClk(H) | nReverseRequest(L),
			  FAST_TIMEOUT, "ECP 2"))
		goto abort;

	SETC(0, DIRECTION_IN | ACKINT_DISABLE,
	     PeriphClk(H) | PeriphAck(H) | nAckReverse(L) | nPeriphRequest(L) | Xflag(H));

	return 0;

abort:
	SETC(0, DIRECTION_IN | ACKINT_DISABLE,
	     PeriphClk(H) | PeriphAck(H) | nAckReverse(L) | nPeriphRequest(L) | Xflag(H));
	pemu_log_string(pe->log_low, "!!!! ABORT IN REVERSE MODE !!!!\n");
	return -1;
}

static int
pemu_send_byte(struct pemu *pe, unsigned data, int eof)
{
	switch (pe->mode & ~4) {
	case 0:
		return (pemu_send_nibble(pe, data & 0x0f, 0) ||
			pemu_send_nibble(pe, (data >> 4) & 0x0f, eof));
	case 0x10:
		return pemu_send_ecp_byte(pe, data, eof);
	default:
		pemu_log_string(pe->log_low, "UNEXPECTED MODE\n");
		return -1;
	}
}

int
pemu_send(struct pemu *pe, const unsigned char *buf, size_t nbytes, int eof)
{
	for (; nbytes--; buf++) {
		if (pemu_send_byte(pe, *buf, eof && nbytes == 0)) {
			pe->mode = -1;
			pemu_log_string(pe->log_low, "SEND ERROR\n");
			return -1 - nbytes;
		}
		log_len = sprintf(log_line, "<- %2.2x\n", *buf);
		pemu_log(pe->log_low, log_line, log_len);
	}
	return 0;
}

int
pemu_send_id(struct pemu *pe, const char *id_string)
{
	short slen = strlen(id_string);
	short blen = slen + 2;
	char lenbuf[2] = {blen >> 8, blen & 0xff};

	pemu_log_string(pe->log_low, "NIBBLE ID\n");
	if (pemu_send(pe, lenbuf, 2, 0) ||
	    pemu_send(pe, id_string, slen, 1))
		return -1;
	return 0;
}

static void
pemu_1284_mode(struct pemu *pe)
{
	int mode, mode_available, enter_1284, data_available;

pemu_log_ppstatus(pe);

	SET( PtrClk(L) | PtrBusy(L) | nDataAvail(H) | AckDataReq(H) | Xflag(H) );

	pe->mode = inb(DATA(0));
	pemu_log_int(pe->log_low, "IEEE 1248 mode %2.2x: ", pe->mode);

	if (pemu_wait_for(pe, HostBusy(H) | HostClk(H),
			  Active1284(H), TIMEOUT, "INIT 1284"))
		goto abort;

	mode = pe->mode & ~ID_MODE_FLAG;

	switch (mode) {
	case NIBBLE_MODE:
	case ECP_MODE:
		mode_available = 1;

		switch (pe->negotiate ? pe->negotiate(pe, pe->mode) :
			((pe->mode & ID_MODE_FLAG) && pe->id_string) ? ACCEPT_1284 :
			pe->send ? ACCEPT_1284 : NO_DATA_1284) {

		case ACCEPT_1284:
			pemu_log_string(pe->log_low, "ACCEPT\n");
			data_available = 1;
			enter_1284 = 1;
			break;

		case NO_DATA_1284:
			pemu_log_string(pe->log_low, "NO DATA\n");
			data_available = 0;
			enter_1284 = (mode != NIBBLE_MODE);
			break;

		default:
			pemu_log_string(pe->log_low, "REJECT\n");
			mode_available = 0;
			data_available = 0;
			enter_1284 = 0;
			break;
		}
		break;

	default:
			pemu_log_string(pe->log_low, "REJECT UNKNOWN MODE\n");
			mode_available = 0;
			data_available = 0;
			enter_1284 = 0;
	}

	if (data_available)
		SET( nDataAvail(L) | AckDataReq(L) | Xflag(H) );
	else if (mode_available)
		SET( nDataAvail(H) | AckDataReq(L) | Xflag(H) );
	else
		SET( nDataAvail(H) | AckDataReq(L) | Xflag(L) );
	SET(PtrClk(H));
	/* negotiation done */

	if (enter_1284) {
		if (pemu_wait_for(pe, HostBusy(L),
				  Active1284(H) , TIMEOUT, "SETUP 1284"))
			goto abort;

		switch (mode) {
		case ECP_MODE:
			SET( nAckReverse(H) );
			SET( nPeriphRequest(L) );

			if (pemu_wait_for(pe, nReverseRequest(L),
					  Active1284(H) , TIMEOUT, "ECP REVERSE REQUEST"))
				goto abort;

			SET( nAckReverse(L) );
			SET( PeriphAck(H) ); /* normal data, no address */
			break;
		}

		/* send data */
		if (data_available) {
			pemu_log_string(pe->log_low, "SEND\n");
			if ((pe->mode & ID_MODE_FLAG) && pe->id_string)
				pemu_send_id(pe, pe->id_string);
			else if (pe->send)
				pe->send(pe);
		}

		/* cleanup reverse mode */
		switch (mode) {
		case ECP_MODE:
			SET( nPeriphRequest(H) );

#if 1
			SET( PeriphClk(L) );
			SET( PeriphClk(H) );
#endif
			if (pemu_wait_for(pe, nReverseRequest(H),
					  Active1284(H) | HostAck(L),
					  TIMEOUT, "ECP FORWARD REQUEST"))
				goto abort;

			/* Data port 0 must be tri-stated at least now! */
			/* (allready done in pemu_send_ecp_byte)        */

			SET( PeriphAck(L) | PeriphClk(H) );
			SET( nAckReverse(H) );
			break;
		}
	}

	/* terminate 1284 mode */
	if (pemu_wait_for(pe, Active1284(L) | HostBusy(H),
			  0, TIMEOUT, "TERMINATE 1284 1"))
		goto abort;

	if (enter_1284) {
		SET( PtrBusy(H) | nDataAvail(H) | AckDataReq(L) );
		SET( PtrClk(L) | Xflag(X) );

		pemu_wait_for(pe, HostBusy(L),
			      0 , TIMEOUT, "TERMINATE 1284 2");
	    }

	/* prepare for compatibility mode */
	SET( Busy(H) | PError(L) | Select(H) | nFault(H) );
	SET( PtrClk(H) );

	pemu_wait_for(pe, HostBusy(H),
		      0 , TIMEOUT, "TERMINATE 1284 3");

	return;

abort:
	pemu_reset_ports();
	pemu_wait_for(pe, Active1284(L),
		      0 , 0, "ABORT 1284");
}

static void
pemu_run(struct pemu* pe)
{
	pemu_log_string(pe->log_low, "START\n");

	while (1) {
		/* SPP Data */
		if (CHECK( Active1284(L) | nStrobe(H) ))
			pemu_spp_mode(pe);

		/* 1248 Negotiation */
		if (CHECK( Active1284(H) | HostBusy(L) )) {
			pemu_1284_mode(pe);
			pemu_log_string(pe->log_low, "FORWARD\n");
		}

		/* Reschedule */
		if(pemu_reschedule(pe))
			break;
	}

	SET(Busy(H));

	pemu_log_string(pe->log_low, "END\n");
}

static int
pemu_reschedule(struct pemu *pe)
{
	if (trace) {
		static short old_wstatus = 0xffff;
		unsigned short new_wstatus = inw(STATUS(0));
		if (new_wstatus != old_wstatus) {
			old_wstatus = new_wstatus;
			log_len = sprintf(log_line, "#%4.4x %8.8x\n", new_wstatus, jiffies);
			pemu_log(pe->log_low, log_line, log_len);
		}
	}

	if (jiffies > pe->busy_timeout || jiffies < pe->sleep_timeout) {
		if (pe->busy_timeout) {
			pe->busy_timeout = 0;
			pemu_log_string(pe->log_low, "BUSY TIMEOUT\n");
		}

		while (buffer_full.counter || current->need_resched || jiffies < pe->sleep_timeout) {
			schedule();
#if 0
			if (!buffer_full.counter)
				pemu_log_string(pe->log_low, "SCHEDULE\n");
#endif
			if (signal_pending(current)) {
				pemu_log_string(pe->log_low, "INTERRUPT\n");
				return -EINTR;
			}
		}
	}

	if (signal_pending(current)) {
		pemu_log_string(pe->log_low, "INTERRUPT\n");
		return -EINTR;
	}

	return 0;
}

static int
pemu_wait_for(struct pemu *pe, long flags, long abort_flags,
	      unsigned long timeout, const char *abort_string)
{
	unsigned long time_limit = jiffies + timeout;

	while (1) {
		if (!CHECK(abort_flags)) {
			pemu_log_string(pe->log_low, "ABORTING  ");
			pemu_log_string(pe->log_low, abort_string);
			pemu_log_string(pe->log_low, "\n");
			pemu_log_ppstatus(pe);
			pe->mode = -1;
			return 1;
		}

		if (CHECK(flags))
			return 0;

		if (timeout) {
			if (jiffies > time_limit) {
				pemu_log_string(pe->log_low, "TIMEOUT  ");
				pemu_log_string(pe->log_low, abort_string);
				pemu_log_string(pe->log_low, "\n");
				pemu_log_ppstatus(pe);
				return 2;
			}
		} else {
			if (pemu_reschedule(pe))
				return -EINTR;
		}
	}
}

static void
pemu_intr_func(int irq, void *dev_id, struct pt_regs *regs)
{
	irq_count++;
}

/* BUFFER */

static struct pemu_buffer*
new_pemu_buffer(struct pemu_log *log)
{
	struct pemu_buffer *buffer = kmalloc(sizeof(struct pemu_buffer), GFP_KERNEL);
	if (!buffer) return NULL;

	buffer->prev = NULL;
	buffer->next = NULL;
	buffer->pos  = 0;

	{
		unsigned long flags;
		spin_lock_irqsave (&log->buffer_lock, flags);
		buffer->next = log->buffer_list;
		log->buffer_list = buffer;
		if (buffer->next)
			buffer->next->prev = buffer;
		spin_unlock_irqrestore (&log->buffer_lock, flags);
	}

	return buffer;
};

static void
del_pemu_buffer(struct pemu_log *log, struct pemu_buffer *buffer)
{
	{
		unsigned long flags;
		spin_lock_irqsave (log->buffer_lock, flags);
		if (buffer->next)
			buffer->next->prev = buffer->prev;
		if (buffer->prev)
			buffer->prev->next = buffer->next;
		else
			log->buffer_list = buffer->next;
		spin_unlock_irqrestore (log->buffer_lock, flags);
	}
	kfree(buffer);
}

int
pemu_log(struct pemu_log *log, const char * buf, size_t nbytes)
{
	struct pemu_buffer *buffer;
	long max_free = BUFFERSIZE;

	if (nbytes > BUFFERSIZE / 8) {
		buf = "\nLOG OVERFLOW\n";
		nbytes = strlen(buf);
	}

	{
		unsigned long flags;
		spin_lock_irqsave (log->buffer_lock, flags);

		for (buffer = log->buffer_list; buffer; buffer = buffer->next)
			if (BUFFERSIZE - buffer->pos < max_free)
				max_free = BUFFERSIZE - buffer->pos;

		if (nbytes > max_free) {
			char overflow[] = "\nBUFFER OVERFLOW\n";
			for (buffer = log->buffer_list; buffer; buffer = buffer->next) {
				buffer->pos = strlen(overflow);
				memcpy(buffer->buf, overflow, buffer->pos);
			}
			spin_unlock_irqrestore (log->buffer_lock, flags);
			return -1;
		}

		for (buffer = log->buffer_list; buffer; buffer = buffer->next) {
			memcpy(buffer->buf + buffer->pos, buf, nbytes);
			buffer->pos += nbytes;

			if (buffer->pos > BUFFERSIZE / 2 &&
			    !test_and_set_bit(0, &buffer->buffer_full))
				atomic_add(1, &buffer_full);
		}

		spin_unlock_irqrestore (log->buffer_lock, flags);
	}
	return 0;
}

int
pemu_log_string(struct pemu_log *log, const char * string)
{
	return pemu_log(log, string, strlen(string));
}

int
pemu_log_int(struct pemu_log *log, const char * format, int val)
{
	char tmpbuf[1000];
	int len = sprintf(tmpbuf, format, val);
	return pemu_log(log, tmpbuf, len);
}

static int
pemu_buffer_read(struct pemu_buffer *buffer, char * buf, size_t nbytes, int block)
{
	if (!nbytes)
		return 0;

	while(!buffer->pos) {
		if (!block)
			return -EAGAIN;
		if (current->need_resched)
			schedule();
		if (signal_pending(current))
			return  -EINTR;	
	}

	if (nbytes > buffer->pos)
		nbytes = buffer->pos;
	if (!nbytes)
		return 0;

	memcpy(buf, buffer->buf, nbytes);
	buffer->pos -= nbytes;
	if (buffer->pos)
		memmove(buffer->buf, buffer->buf + nbytes, buffer->pos);

	if (buffer->pos < BUFFERSIZE / 4 &&
	    test_and_clear_bit(0, &buffer->buffer_full))
		atomic_sub(1, &buffer_full);

	return nbytes;
}

static int
run_open(struct inode * inode, struct file * file)
{
	struct proc_dir_entry *dp = (struct proc_dir_entry *) inode->u.generic_ip;
	struct pemu *pe = (struct pemu*) dp->data;

	if (test_and_set_bit(0, &running))
		return -EBUSY;

	pemu_run(pe);

	running = 0;
	return 0;
}

static int
log_open(struct inode * inode, struct file * file)
{
	struct proc_dir_entry *dp = (struct proc_dir_entry *) inode->u.generic_ip;
	struct pemu_log *log = (struct pemu_log*) dp->data;
	file->private_data = new_pemu_buffer(log);
	return file->private_data ? 0 : -ENOMEM;
}

static int
log_close(struct inode * inode, struct file * file)
{
	struct proc_dir_entry *dp = (struct proc_dir_entry *) inode->u.generic_ip;
	struct pemu_log *log = (struct pemu_log*) dp->data;
	del_pemu_buffer(log, (struct pemu_buffer*) file->private_data);
	return 0;
}

static ssize_t
log_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
{
	return pemu_buffer_read((struct pemu_buffer*) file->private_data,
				buf, nbytes, !(file->f_flags & O_NONBLOCK));

}

static long long
no_lseek(struct file * file, long long offset, int orig)
{
	return -EPIPE;
}

static ssize_t
no_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
{
	return -EIO;
}

static ssize_t
no_write(struct file * file, const char * buffer,
	  size_t count, loff_t *ppos)
{
	return -EIO;
}

static struct file_operations log_file_operations = {
	no_lseek,	/* lseek	*/
	log_read,	/* read		*/
	no_write,	/* write	*/
	NULL,		/* readdir	*/
	NULL,		/* poll		*/
	NULL,		/* ioctl	*/
	NULL,		/* mmap		*/
	log_open,	/* special open code	*/
	NULL,		/* flush	*/
	log_close,	/* special release code	*/
	NULL		/* can't fsync	*/
};

static struct inode_operations log_file_inode_operations = {
	&log_file_operations,  /* default log file-ops */
	NULL,		/* create	*/
	NULL,		/* lookup	*/
	NULL,		/* link		*/
	NULL,		/* unlink	*/
	NULL,		/* symlink	*/
	NULL,		/* mkdir	*/
	NULL,		/* rmdir	*/
	NULL,		/* mknod	*/
	NULL,		/* rename	*/
	NULL,		/* readlink	*/
	NULL,		/* follow_link  */
	NULL,		/* readpage	*/
	NULL,		/* writepage    */
	NULL,		/* bmap	        */
	NULL,		/* truncate	*/
	NULL		/* permission   */
};

static struct file_operations run_file_operations = {
	no_lseek,	/* lseek	*/
	no_read,	/* read		*/
	no_write,	/* write	*/
	NULL,		/* readdir	*/
	NULL,		/* poll		*/
	NULL,		/* ioctl	*/
	NULL,		/* mmap		*/
	run_open,	/* special open code	*/
	NULL,		/* flush	*/
	NULL,		/* no special release code	*/
	NULL		/* can't fsync	*/
};

static struct inode_operations run_file_inode_operations = {
	&run_file_operations,  /* default run file-ops */
	NULL,		/* create	*/
	NULL,		/* lookup	*/
	NULL,		/* link		*/
	NULL,		/* unlink	*/
	NULL,		/* symlink	*/
	NULL,		/* mkdir	*/
	NULL,		/* rmdir	*/
	NULL,		/* mknod	*/
	NULL,		/* rename	*/
	NULL,		/* readlink	*/
	NULL,		/* follow_link  */
	NULL,		/* readpage	*/
	NULL,		/* writepage    */
	NULL,		/* bmap	        */
	NULL,		/* truncate	*/
	NULL		/* permission   */
};

static void modcount(struct inode *inode, int fill)
{
#ifdef MODULE
	if (fill) {
		MOD_INC_USE_COUNT;
	} else {
		MOD_DEC_USE_COUNT;
	}
#endif
}

struct proc_dir_entry *
new_proc_entry(const char *name, mode_t mode,
	       struct proc_dir_entry *parent,
	       unsigned short ino,
	       struct inode_operations *ops)
{
	struct proc_dir_entry *ent;

	ent = kmalloc(sizeof(struct proc_dir_entry), GFP_KERNEL);
	if (!ent)
		return NULL;

	memset(ent, 0, sizeof(struct proc_dir_entry));
	
	if (mode == S_IFDIR)
		mode |= S_IRUGO | S_IXUGO;
	else if (mode == 0)
		mode = S_IFREG | S_IRUGO;

	ent->low_ino = ino;
	ent->name = name;
	ent->namelen = strlen(name);
	ent->mode = mode;
	ent->ops = ops;

	if (S_ISDIR(mode))
	{
		ent->fill_inode = modcount;
		ent->nlink = 2;
	} else
		ent->nlink = 1;

	proc_register(parent, ent);
	
	return ent;
}

void
destroy_proc_entry(struct proc_dir_entry *root, 
		   struct proc_dir_entry **d)
{
	proc_unregister(root, (*d)->low_ino);
	kfree(*d);
	*d = NULL;
}

struct pemu_log*
new_pemu_log(struct pemu* pe, const char *name) {
	struct pemu_log *log = kmalloc(sizeof(struct pemu_log), GFP_KERNEL);
	if (!log)
		return NULL;


	spin_lock_init(log->buffer_lock);
	log->buffer_list = NULL;
	log->pe = pe;

	log->proc_log = new_proc_entry(name, S_IFREG | S_IRUGO, pe->proc_dir, 0, &log_file_inode_operations);
	log->proc_log->data = log;

	return log;
}

void
del_pemu_log(struct pemu_log *log) {
	if (log) {
		if (log->proc_log)
			destroy_proc_entry(log->pe->proc_dir, &log->proc_log);
		kfree(log);
	}
}

struct pemu*
new_pemu(const char *dirname)
{
	struct pemu *pe;

	pemu_reset_ports();
	request_irq(IRQ0, pemu_intr_func, 0, dirname ? : "pemu", NULL);

	pe = kmalloc(sizeof(struct pemu), GFP_KERNEL);
	if (!pe)
		return NULL;

	pe->mode         = 0;
	pe->busy_timeout = 0;
	pe->sleep_timeout = 0;
	pe->id_string    = NULL;
	pe->negotiate    = NULL;
	pe->receive      = NULL;
	pe->send         = NULL;
	pe->data         = NULL;

	pe->proc_dir = new_proc_entry(dirname ? : "pemu", S_IFDIR, &proc_root, 0, NULL);
	if (!pe->proc_dir) {
		kfree(pe);
		return NULL;
	}

	pe->proc_status = new_proc_entry("ppstatus", S_IFREG | S_IRUGO, pe->proc_dir, 0, NULL);
	pe->proc_status->read_proc = ppstatus_read_proc;
	pe->proc_status->data = pe;

	pe->log_low = new_pemu_log(pe, "log_low");
	pe->log_high = new_pemu_log(pe, "log_high");

	running = 1;
	pe->proc_run = new_proc_entry("run", S_IFREG | S_IRUGO | S_IXUGO , pe->proc_dir, 0, &run_file_inode_operations);
	pe->proc_run->data = pe;

	running = 0;

	return(pe);
}

void
del_pemu(struct pemu *pe)
{
	free_irq(IRQ0, NULL);

	if (pe->proc_run)
		destroy_proc_entry(pe->proc_dir, &pe->proc_run);
	del_pemu_log(pe->log_low);
	if (pe->proc_status)
		destroy_proc_entry(pe->proc_dir, &pe->proc_status);
	if (pe->proc_dir)
		destroy_proc_entry(&proc_root, &pe->proc_dir);
	kfree(pe);
}
