#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/parport.h>
#include <asm/spinlock.h>
#include <linux/delay.h>

#define PPS_UNINITIALIZED 0
#define PPS_PORT_OWNER    1
#define PPS_LURK          2
#define PPS_CLEANUP       4

#define DIRECTION (1 << 5)
#define DIRECTION_REVERSE (1 << 5)
#define MODE (7 << 5)
#define MODE_PS2 (1 << 5)

struct pps_struct {
	struct pardevice *dev;
	struct pps_struct *next;
	struct pps_struct *prev;
	int state;
	struct proc_dir_entry *proc_slave;
	struct proc_dir_entry *proc_trace;
};

struct pps_struct *pps_list = NULL;
struct pps_struct *pps_list_tail = NULL;
spinlock_t pps_lock = SPIN_LOCK_UNLOCKED;

#define	P_DATA			0
#define	P_STATUS		1
#define	P_CONTROL		2
#define P_EPPADDR		3
#define P_EPPDATA		4
#define P_ECPA			0x400
#define P_ECPB			0x401
#define P_ECR			0x402

#define	P_GET(p, port)		inb(p->base + port)

struct inode_operations pps_file_inode_operations;

int
dump_status(struct parport *pp, char *buf)
{
	int data = pp->ops->read_data(pp);
	int status = pp->ops->read_status(pp);
	int control = pp->ops->read_control(pp);
	int econtrol = pp->ops->read_econtrol(pp);
	return sprintf(buf, "%8.8x %2.2x %2.2x %2.2x %2.2x\n",
		       jiffies, econtrol, control, status, data);
}

static int
pps_open(struct inode * inode, struct file * file)
{
	return 0;
}

static int
pps_close(struct inode * inode, struct file * file)
{
	return 0;
}

static ssize_t
pps_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
{
	struct inode * inode = file->f_dentry->d_inode;
	struct proc_dir_entry * dp = (struct proc_dir_entry *) inode->u.generic_ip;
	struct pps_struct *pps = (struct pps_struct*) dp->data;
	struct parport *pp = pps->dev->port;
	int data;
	int len = 0;

	if (nbytes < 1)
		return 0;

	while (!(pp->ops->read_status(pp) & 0x40))
	{
		if (current->need_resched)
			schedule();
		if (signal_pending(current))
			return  -EINTR;
	}
	pp->ops->frob_control(pp, 2, 2);
	pp->ops->frob_control(pp, 1, 0);
	while (pp->ops->read_status(pp) & 0x40)
	{
		if (current->need_resched)
		{
			pp->ops->frob_control(pp, 2, 0);
			udelay(1);
			if (!(pp->ops->read_status(pp) & 0x40))
				break;
			schedule();
			pp->ops->frob_control(pp, 2, 2);
		}
		if (signal_pending(current))
		{
			pp->ops->frob_control(pp, 2, 0);
			return  -EINTR;
		}
	}
	pp->ops->frob_control(pp, 2, 0);
	data = pp->ops->read_data(pp);
	pp->ops->frob_control(pp, 1, 1);

	*buf++ = data;
	len++;
	nbytes--;


#if 0
	len += sprintf(page + len, "%2.2x (%c)\n", data, data);
#endif

	return len;
}

static int
trace_read_proc(char *page, char **start, off_t off,
	      int count, int *eof, void *handle)
{
	struct pps_struct *pps = (struct pps_struct*) handle;
	struct parport *pp = pps->dev->port;
	int len;
	int data = pp->ops->read_data(pp);
	int status = pp->ops->read_status(pp);
	int control = pp->ops->read_control(pp);
	int econtrol = pp->ops->read_econtrol(pp);


	len = sprintf(page, "%8.8x %2.2x %2.2x %2.2x %2.2x\n",
		      jiffies, econtrol, control, status, data);

	off = off % len;
	len -= off;
	*start = page + off;

	return len <= count ? len : count;
}

static struct proc_dir_entry *
new_proc_entry(const char *name, mode_t mode,
	       struct proc_dir_entry *parent,
	       unsigned short ino,
	       struct inode_operations *ops,
	       struct parport *p)
{
	struct proc_dir_entry *ent;

	ent = kmalloc(sizeof(struct proc_dir_entry), GFP_KERNEL);
	if (!ent)
		return NULL;

	memset(ent, 0, sizeof(struct proc_dir_entry));
	
	if (mode == S_IFDIR)
		mode |= S_IRUGO | S_IXUGO;
	else if (mode == 0)
		mode = S_IFREG | S_IRUGO;

	ent->low_ino = ino;
	ent->name = name;
	ent->namelen = strlen(name);
	ent->mode = mode;
	ent->ops = ops;

	if (S_ISDIR(mode))
	{
		if (p && p->ops)
			ent->fill_inode = p->ops->fill_inode;
		ent->nlink = 2;
	} else
		ent->nlink = 1;

	proc_register(parent, ent);
	
	return ent;
}

static inline void
destroy_proc_entry(struct proc_dir_entry *root, 
		   struct proc_dir_entry **d)
{
	proc_unregister(root, (*d)->low_ino);
	kfree(*d);
	*d = NULL;
}

static void
pps_initialize_port(struct pps_struct *pps)
{
	struct parport *pp = pps->dev->port;
	printk(KERN_DEBUG "pps: Initialize\n");
	pps->state = PPS_PORT_OWNER | PPS_LURK;

	pp->ops->frob_econtrol(pp, MODE, MODE_PS2);
	pp->ops->frob_control(pp, DIRECTION, DIRECTION_REVERSE);
	pp->ops->frob_control(pp, 0x10, 0);
	pp->ops->frob_control(pp, 0x0f, 0);
}

static void
pps_wakeup(void *handle)
{
	struct pps_struct *pps = (struct pps_struct*) handle;
	printk(KERN_DEBUG "pps: wakeup\n");

	if (pps->state & (PPS_PORT_OWNER | PPS_CLEANUP)) {
		printk(KERN_DEBUG "pps: why am I being woken up?\n");
		return;
	}

	if (!parport_claim(pps->dev))
	{
		if (pps->state == PPS_UNINITIALIZED)
			pps_initialize_port(pps);
		pps->state |= PPS_PORT_OWNER;
		printk(KERN_DEBUG "pps: registered\n");
	}
}

static int
pps_preempt(void *handle)
{
	struct pps_struct *pps = (struct pps_struct*) handle;
	printk(KERN_DEBUG "pps: preempt\n");

	if (! (pps->state & PPS_PORT_OWNER))
		printk(KERN_DEBUG "pps: preempt: not owner\n");

	if (! (pps->state & PPS_LURK))
		return 1;

return 1;


	pps->state &= ~PPS_PORT_OWNER;
	return 0;
}

static void
pps_interrupt(int irq, void *handle, struct pt_regs * regs)
{
#if 0
	struct pps_struct *pps = (struct pps_struct*) handle;
	struct parport *pp = pps->dev->port;
#endif
}

struct pps_struct*
pps_register(struct parport *port)
{
	struct pardevice *dev;
	struct pps_struct *pps;

	pps = kmalloc(sizeof(struct pps_struct), GFP_KERNEL);
	if (!pps) {
		printk(KERN_ERR "pps: memory squeeze\n");
		return NULL;
	}
	
	dev = parport_register_device(port, "slave", 
				      pps_preempt, pps_wakeup,
				      pps_interrupt,
				      PARPORT_DEV_LURK,
				      (void *) pps);
	if (dev == NULL)
	{
		kfree(pps);
		printk(KERN_ERR "pps: could not register device\n");
		return NULL;
	}


	pps->state = PPS_UNINITIALIZED;
	pps->dev = dev;
	
	pps->proc_trace = new_proc_entry("trace", S_IFREG | S_IRUGO, dev->port->pdir.entry, 0, NULL, dev->port);
	pps->proc_trace->read_proc = trace_read_proc;
	pps->proc_trace->data = pps;

	pps->proc_slave = new_proc_entry("slave", S_IFREG | S_IRUGO, dev->port->pdir.entry, 0, &pps_file_inode_operations, dev->port);
	pps->proc_slave->data = pps;

	pps->next = NULL;
	{
		unsigned long flags;
		spin_lock_irqsave (&pps_lock, flags);
		pps->prev = pps_list_tail;
		if (pps_list_tail)
			pps_list_tail->next = pps;
		else
			pps_list = pps;
		pps_list_tail = pps;
		spin_unlock_irqrestore (&pps_lock, flags);
	}

	pps_wakeup((void*) pps);

	return pps;
}

void
pps_unregister(struct pps_struct *pps)
{
	{
		unsigned long flags;
		spin_lock_irqsave (&pps_lock, flags);
		if (pps->state & PPS_CLEANUP)
			return;
		pps->state |= PPS_CLEANUP;
		if (pps->next)
			pps->next->prev = pps->prev;
		else
			pps_list_tail = pps->prev;
		if (pps->prev)
			pps->prev->next = pps->next;
		else
			pps_list = pps->next;
		spin_unlock_irqrestore (&pps_lock, flags);
	}
	printk(KERN_DEBUG "pps: Cleanup\n");

	destroy_proc_entry(pps->dev->port->pdir.entry, &pps->proc_slave);
	destroy_proc_entry(pps->dev->port->pdir.entry, &pps->proc_trace);

	pps->state &= ~PPS_LURK;
	if (pps->state & PPS_PORT_OWNER)
		parport_release(pps->dev);
	parport_unregister_device(pps->dev);

	kfree(pps);
}

int
pps_init(void)
{
	struct parport *port;

	for (port = parport_enumerate(); port; port = port->next)
		pps_register(port);

	return 0;
}


#ifdef MODULE
int
init_module(void)
{
	return pps_init();
}

void
cleanup_module(void)
{
	while (pps_list)
		pps_unregister(pps_list);
}

#if 0
EXPORT_SYMBOL(pps_register);
EXPORT_SYMBOL(pps_unregister);
#endif
#endif

static ssize_t
pps_write(struct file * file, const char * buffer,
	  size_t count, loff_t *ppos)
{
	return -EIO;
}

static long long
pps_lseek(struct file * file, long long offset, int orig)
{
	return -EPIPE;
}

static struct file_operations pps_file_operations = {
    pps_lseek,	/* lseek   */
    pps_read,	/* read	   */
    pps_write,	/* write   */
    NULL,		/* readdir */
    NULL,		/* poll    */
    NULL,		/* ioctl   */
    NULL,		/* mmap	   */
    pps_open,		/* no special open code	   */
    NULL,		/* flush */
    pps_close,		/* no special release code */
    NULL		/* can't fsync */
};

struct inode_operations pps_file_inode_operations = {
	&pps_file_operations,  /* default pps file-ops */
	NULL,		/* create	*/
	NULL,		/* lookup	*/
    NULL,		/* link		*/
    NULL,		/* unlink	*/
    NULL,		/* symlink	*/
    NULL,		/* mkdir	*/
    NULL,	    /* rmdir	   */
    NULL,	    /* mknod	   */
    NULL,	    /* rename	   */
    NULL,	    /* readlink	   */
    NULL,	    /* follow_link */
    NULL,	    /* readpage	   */
    NULL,	    /* writepage   */
    NULL,	    /* bmap	   */
    NULL,	    /* truncate	   */
    NULL	    /* permission  */
};
