#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/parport.h>
#include <asm/uaccess.h>
#include <asm/spinlock.h>
#include <linux/delay.h>

#if 1
#define MAX_TIME    0x4000000
#else
#define MAX_TIME     0x400000
#endif
#define MAX_RECORDS 0x800000
#define RECORD_FORMAT "%4.4x %4.4x\n"
#define RECORD_LEN    10


#define PPS_UNINITIALIZED 0
#define PPS_PORT_OWNER    1
#define PPS_LURK          2
#define PPS_CLEANUP       4

#define DIRECTION (1 << 5)
#define DIRECTION_REVERSE (1 << 5)
#define MODE (7 << 5)
#define MODE_PS2 (1 << 5)

struct record_struct {
	unsigned short time;
	unsigned short wstatus;
};

struct pps_struct {
	struct pardevice *dev;
	struct pps_struct *next;
	struct pps_struct *prev;
	int state;
	struct proc_dir_entry *proc_trace;
	struct proc_dir_entry *proc_status;
	struct record_struct *data;
	long len;
};

struct pps_struct *pps_list = NULL;
struct pps_struct *pps_list_tail = NULL;
spinlock_t pps_lock = SPIN_LOCK_UNLOCKED;

#define	P_DATA			0
#define	P_STATUS		1
#define	P_CONTROL		2
#define P_EPPADDR		3
#define P_EPPDATA		4
#define P_ECPA			0x400
#define P_ECPB			0x401
#define P_ECR			0x402

#define	P_GET(p, port)		inb(p->base + port)

struct inode_operations pps_file_inode_operations;

static int
pps_open(struct inode * inode, struct file * file)
{
	struct proc_dir_entry * dp = (struct proc_dir_entry *) inode->u.generic_ip;
	struct pps_struct *pps = (struct pps_struct*) dp->data;

	if (pps->data)
		return -EACCES;

	pps->data = (struct record_struct*) vmalloc(MAX_RECORDS * sizeof (struct record_struct));
	if (!pps->data)
		return -ENOMEM;

	return 0;
}

static int
pps_close(struct inode * inode, struct file * file)
{
	struct proc_dir_entry * dp = (struct proc_dir_entry *) inode->u.generic_ip;
	struct pps_struct *pps = (struct pps_struct*) dp->data;

	if (pps->data)
		vfree(pps->data);
	pps->data = NULL;

	return 0;
}

static ssize_t
pps_read(struct file * file, char * buf, size_t nbytes, loff_t *ppos)
{
	char record[RECORD_LEN+10];
	struct inode * inode = file->f_dentry->d_inode;
	struct proc_dir_entry * dp = (struct proc_dir_entry *) inode->u.generic_ip;
	struct pps_struct *pps = (struct pps_struct*) dp->data;
	struct parport *pp = pps->dev->port;
	unsigned long base = pp->base;
	unsigned short wstatus = 0xffff;
	unsigned long time, time_end;
	struct record_struct *data = pps->data, *data_end;
	long len = 0;

	if (!*ppos) {
		data_end = data + MAX_RECORDS;
		time = 0;
		time_end = MAX_TIME;

		while (data < data_end && time++ < time_end)
		{
			unsigned short newstatus = inw(base);
			int incr = newstatus == wstatus ? 0 : 1;
			data->time = time & 0xffff;
			data->wstatus = wstatus = newstatus;
			data += incr;
		}

		pps->len = data + MAX_RECORDS - data_end;
	}
	data = pps->data;


	if ((long) *ppos / RECORD_LEN >= pps->len) {
		return 0;
#if 0
		*ppos = 0;
		len = sprintf(record, "\n\nBREAK\n\n");
		if(copy_to_user(buf, record, len))
			return -EIO;
		return len;
#endif
	}

	while(nbytes > 0 && (long) *ppos / RECORD_LEN < pps->len)
	{
		long doffset  = (long) *ppos / RECORD_LEN;
		long roffset  = (long) *ppos % RECORD_LEN;
		unsigned short time    = doffset ? data[doffset-1].time : 0;
		unsigned short wstatus = data[doffset].wstatus;
		int rlen      = RECORD_LEN - roffset;
		if (rlen > nbytes)
			rlen = nbytes;

		sprintf(record, RECORD_FORMAT, time, wstatus);

		if(copy_to_user(buf, record + roffset, rlen))
			return -EIO;

		nbytes -= rlen;
		len    += rlen;
		*ppos  += rlen;
		buf    += rlen;
	}
	return len;
}

static int
status_read_proc(char *page, char **start, off_t off,
	      int count, int *eof, void *handle)
{
	struct pps_struct *pps = (struct pps_struct*) handle;
	struct parport *pp = pps->dev->port;
	int base = pp->base;
	int len;
	int wstatus = inw(base);
	int data = pp->ops->read_data(pp);
	int status = pp->ops->read_status(pp);
	int control = pp->ops->read_control(pp);
	int econtrol = pp->ops->read_econtrol(pp);


	len = sprintf(page, "%8.8x %2.2x %2.2x %2.2x %2.2x %4.4x\n",
		      jiffies, econtrol, control, status, data, wstatus);

	*start = 0;

	return len <= count ? len : count;
}

static struct proc_dir_entry *
new_proc_entry(const char *name, mode_t mode,
	       struct proc_dir_entry *parent,
	       unsigned short ino,
	       struct inode_operations *ops,
	       struct parport *p)
{
	struct proc_dir_entry *ent;

	ent = vmalloc(sizeof(struct proc_dir_entry));
	if (!ent)
		return NULL;

	memset(ent, 0, sizeof(struct proc_dir_entry));
	
	if (mode == S_IFDIR)
		mode |= S_IRUGO | S_IXUGO;
	else if (mode == 0)
		mode = S_IFREG | S_IRUGO;

	ent->low_ino = ino;
	ent->name = name;
	ent->namelen = strlen(name);
	ent->mode = mode;
	ent->ops = ops;

	if (S_ISDIR(mode))
	{
		if (p && p->ops)
			ent->fill_inode = p->ops->fill_inode;
		ent->nlink = 2;
	} else
		ent->nlink = 1;

	proc_register(parent, ent);
	
	return ent;
}

static inline void
destroy_proc_entry(struct proc_dir_entry *root, 
		   struct proc_dir_entry **d)
{
	proc_unregister(root, (*d)->low_ino);
	vfree(*d);
	*d = NULL;
}

static void
pps_initialize_port(struct pps_struct *pps)
{
	struct parport *pp = pps->dev->port;
	printk(KERN_DEBUG "pps: Initialize\n");
	pps->state = PPS_PORT_OWNER | PPS_LURK;

	pp->ops->frob_econtrol(pp, MODE, MODE_PS2);
	pp->ops->frob_control(pp, DIRECTION, DIRECTION_REVERSE);
	pp->ops->frob_control(pp, 0x10, 0);
	pp->ops->frob_control(pp, 0x0f, 0);
}

static void
pps_wakeup(void *handle)
{
	struct pps_struct *pps = (struct pps_struct*) handle;
	printk(KERN_DEBUG "pps: wakeup\n");

	if (pps->state & (PPS_PORT_OWNER | PPS_CLEANUP)) {
		printk(KERN_DEBUG "pps: why am I being woken up?\n");
		return;
	}

	if (!parport_claim(pps->dev))
	{
		if (pps->state == PPS_UNINITIALIZED)
			pps_initialize_port(pps);
		pps->state |= PPS_PORT_OWNER;
		printk(KERN_DEBUG "pps: registered\n");
	}
}

static int
pps_preempt(void *handle)
{
	struct pps_struct *pps = (struct pps_struct*) handle;
	printk(KERN_DEBUG "pps: preempt\n");

	if (! (pps->state & PPS_PORT_OWNER))
		printk(KERN_DEBUG "pps: preempt: not owner\n");

	if (! (pps->state & PPS_LURK))
		return 1;

	pps->state &= ~PPS_PORT_OWNER;
	return 0;
}

static void
pps_interrupt(int irq, void *handle, struct pt_regs * regs)
{
#if 0
	struct pps_struct *pps = (struct pps_struct*) handle;
	struct parport *pp = pps->dev->port;
#endif
}

struct pps_struct*
pps_register(struct parport *port)
{
	struct pardevice *dev;
	struct pps_struct *pps;

	pps = vmalloc(sizeof(struct pps_struct));
	if (!pps) {
		printk(KERN_ERR "pps: memory squeeze\n");
		return NULL;
	}
	
	dev = parport_register_device(port, "trace", 
				      pps_preempt, pps_wakeup,
				      pps_interrupt,
				      PARPORT_DEV_LURK,
				      (void *) pps);
	if (dev == NULL)
	{
		vfree(pps);
		printk(KERN_ERR "pps: could not register device\n");
		return NULL;
	}


	pps->state = PPS_UNINITIALIZED;
	pps->dev = dev;
	pps->data = NULL;
	
	pps->proc_status = new_proc_entry("status", S_IFREG | S_IRUGO, dev->port->pdir.entry, 0, NULL, dev->port);
	pps->proc_status->read_proc = status_read_proc;
	pps->proc_status->data = pps;

	pps->proc_trace = new_proc_entry("trace", S_IFREG | S_IRUGO, dev->port->pdir.entry, 0, &pps_file_inode_operations, dev->port);
	pps->proc_trace->data = pps;

	pps->next = NULL;
	{
		unsigned long flags;
		spin_lock_irqsave (&pps_lock, flags);
		pps->prev = pps_list_tail;
		if (pps_list_tail)
			pps_list_tail->next = pps;
		else
			pps_list = pps;
		pps_list_tail = pps;
		spin_unlock_irqrestore (&pps_lock, flags);
	}

	pps_wakeup((void*) pps);

	return pps;
}

void
pps_unregister(struct pps_struct *pps)
{
	{
		unsigned long flags;
		spin_lock_irqsave (&pps_lock, flags);
		if (pps->state & PPS_CLEANUP)
			return;
		pps->state |= PPS_CLEANUP;
		if (pps->next)
			pps->next->prev = pps->prev;
		else
			pps_list_tail = pps->prev;
		if (pps->prev)
			pps->prev->next = pps->next;
		else
			pps_list = pps->next;
		spin_unlock_irqrestore (&pps_lock, flags);
	}
	printk(KERN_DEBUG "pps: Cleanup\n");

	destroy_proc_entry(pps->dev->port->pdir.entry, &pps->proc_trace);
	destroy_proc_entry(pps->dev->port->pdir.entry, &pps->proc_status);

	pps->state &= ~PPS_LURK;
	if (pps->state & PPS_PORT_OWNER)
		parport_release(pps->dev);
	parport_unregister_device(pps->dev);

	vfree(pps);
}

int
pps_init(void)
{
	struct parport *port;

	for (port = parport_enumerate(); port; port = port->next)
		pps_register(port);

	return 0;
}


#ifdef MODULE
int
init_module(void)
{
	return pps_init();
}

void
cleanup_module(void)
{
	while (pps_list)
		pps_unregister(pps_list);
}

#if 0
EXPORT_SYMBOL(pps_register);
EXPORT_SYMBOL(pps_unregister);
#endif
#endif

static ssize_t
pps_write(struct file * file, const char * buffer,
	  size_t count, loff_t *ppos)
{
	return -EIO;
}

static long long
pps_lseek(struct file * file, long long offset, int orig)
{
	return -EPIPE;
}

static struct file_operations pps_file_operations = {
    pps_lseek,	/* lseek   */
    pps_read,	/* read	   */
    pps_write,	/* write   */
    NULL,		/* readdir */
    NULL,		/* poll    */
    NULL,		/* ioctl   */
    NULL,		/* mmap	   */
    pps_open,		/* no special open code	   */
    NULL,		/* flush */
    pps_close,		/* no special release code */
    NULL		/* can't fsync */
};

struct inode_operations pps_file_inode_operations = {
	&pps_file_operations,  /* default pps file-ops */
	NULL,		/* create	*/
	NULL,		/* lookup	*/
    NULL,		/* link		*/
    NULL,		/* unlink	*/
    NULL,		/* symlink	*/
    NULL,		/* mkdir	*/
    NULL,	    /* rmdir	   */
    NULL,	    /* mknod	   */
    NULL,	    /* rename	   */
    NULL,	    /* readlink	   */
    NULL,	    /* follow_link */
    NULL,	    /* readpage	   */
    NULL,	    /* writepage   */
    NULL,	    /* bmap	   */
    NULL,	    /* truncate	   */
    NULL	    /* permission  */
};
